//
// Copyright(C) 2001-2005 Simon Howard
//
// You can redistribute and/or modify this program under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation, or any later version. This program is distributed WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
//
// Keep track of and display heads-up splats.
//

#ifndef __SWSPLAT_H__
#define __SWSPLAT_H__

extern void swclearsplats(void);
extern void swdispsplats(void);

extern void swsplatbird(void);
extern void swsplatox(void);
extern void swwindshot(void);

#endif /* #ifndef __SWSPLAT_H__ */
