//
// Copyright(C) 1984-2000 David L. Clark
// Copyright(C) 2001-2005 Simon Howard
//
// You can redistribute and/or modify this program under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation, or any later version. This program is distributed WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//

#ifndef __SWCOLLSN_H__
#define __SWCOLLSN_H__

#include "sw.h"

extern bool CollisionTest(OBJECTS *ob1, OBJECTS *ob2);
extern void swcollsn(void);
extern void scorepln(OBJECTS *ob, obtype_t type);

#endif
